#ifndef N2NCFILTER_H
#define N2NCFILTER_H

// #define FILTER_CONST_CONTINUE 0
// #define FILTER_CONST_RETURN 2
// #define FILTER_CONST_DROP 1


#include "nixsys.h"
// #include "n2nc.h"
#include "blowfish.h"
#include "packetmanager.h"

namespace n2nc {

/**
  \brief Base class for plugin-based traffic filter

  @author fabsoft <fabsoft@gmail.com>
  \note Plugin has to define these factory/cleaner with 
    \code 
extern "C"
static Filter* get_istance(void *argv,int argc);
extern "C"
static int free_istance();
    \endcode
*/
class Filter{
  protected:
// public:
    std::string m_name ;
    std::string m_description ;
    bool m_isenabled ;

    std::string m_filename ;
    void *this_handle;
    
    security::BlowFish::key_t *m_skey ;
    n2nc::PacketManager *m_pkm ;
    
    /* raw information */
    n2nc::net::Socket *m_com_sock ;
    
    /** Each filter has to set its name and description by constructor call.
        \note Use Filter::load_filter to creates a new Filter
     */
    Filter(std::string name,std::string description){
      std::cerr << "loaded filter module: " << name << std::endl ;
      this->m_name = name ;
      this->m_description = description ;
      this->m_com_sock = NULL ;
    }

public:
    /* since plugin is linked alone, i need to define(assing) the enums */
    /** Defines the packet behaviour in the chain */
    enum status_t{
      /** The packet follows the filter chain */
      FILTER_CONTINUE = 0,
      /** The packet exit the filter chain */
      FILTER_RETURN,
      /** The packet is discarded */
      FILTER_DROP
    };

    virtual ~Filter(){
      std::cerr << "UN-loaded filter module: " << this->m_name <<  std::endl ;
    }
    /** Outgoing traffic goes through egress hook. */
    virtual status_t egress(void *inbuf, void *outbuf,size_t inlen,size_t *outlen) = 0 ;
    
    /** Ingoing traffic goes through ingress hook */
    virtual status_t ingress(void *inbuf, void *outbuf,size_t inlen,size_t *outlen) = 0 ;
    
//     virtual status_t egress(void **buf,size_t *len) = 0 ;
//     virtual status_t ingress(void **buf,size_t *len) = 0 ;
    /** Sets the session key. usually 128 bit */
    virtual int setSessionKey(security::BlowFish::key_t *key){ this->m_skey = key ;}
    /** Set the packet manager to the filter, used in certain circumstances */
    virtual int setPKM(n2nc::PacketManager *pkm){this->m_pkm = pkm ;}
    
    /** Prints the filter name and description.. */
    std::string toString(){
      std::ostringstream os;
      std::string s;
      os << "Name: " << this->m_name << " Description: " << this->m_description << " Filename: " << this->m_filename ;   ;
      s =  "Name: " + this->m_name + " Description: " + this->m_description + " Filename: " + this->m_filename  ;
      /* FIXME string */
      return os.str();
    }
    /** Public factory method to istanciates a new Filter module by \a filename */
    static Filter *load_filter(const std::string filename){
      void *handle ;
      Filter *ist;
      typedef Filter* (*get_istance_t)(/*void *argc,int argc*/);
      get_istance_t tmp_get_istance;
//       handle = dlopen(filename.c_str(),/*RTLD_LOCAL|RTLD_NOW*/ RTLD_LAZY );
      handle = dlopen(filename.c_str(),RTLD_NOW );
      if (handle==NULL){ std::cerr << "dlopen fails!:" << dlerror() << std::endl; return NULL ;}
      tmp_get_istance = (get_istance_t)dlsym(handle,"get_istance");
      if (tmp_get_istance ==NULL ) {std::cerr << "dlsym fails!:" << dlerror() << std::endl ; return NULL ;}
      ist = tmp_get_istance() ;
      ist->this_handle = handle ;
      ist->m_filename = filename ;
      return ist ;
    }
    /** Public denstructor method.
    \note The library will be unloaded only if there are no symbols linked against it.
    */
    static int free_filter(Filter *filter){
      void *handle;
      typedef int (*free_istance_t)(Filter *filter);
      free_istance_t tmp_free_istance;
      tmp_free_istance = (free_istance_t)dlsym(filter->this_handle,"free_istance");
      if (tmp_free_istance ==NULL ) {std::cerr << "dlsym fails!:" << dlerror(); exit(1);}
      handle = filter->this_handle ;
      //destructor call
      tmp_free_istance(filter);
      //module unload
      dlclose(handle);
    }



  /** 
    \example filter_ex.cpp Example on how to use the Filter interface.\n
    \link filter_dummy.cpp Filter Implementation. \endlink \n
    \link filter_dummy.h  Filter Header. \endlink

  */

};


}//ns


#endif
