#ifndef N2NC_SYNCTHREAD_H
#define N2NC_SYNCTHREAD_H

#include "nixsys.h"
#include "mutex.h"

namespace n2nc {
namespace sync {

/**
  \brief Base Threads interface

  @author fabsoft <fabsoft@gmail.com>
*/
class Thread{
  
  public:
    
    /** Runs the thread by calling entry_point() routine */
    int run(void *args);
    /** Waits (join) the thread and returns the exit-value */
    int wait(void **retval);
    /** FIXME getRetVal */
    void* getRetVal();
    /** Cancels the thread execution */
    int terminate();
    /** @Returns the thread ID */
    int getID();
    /** Compares thread t. Same of Thread::compare(this, t2) */
    bool isEquals(Thread *t);
    /** @returns the thread name/description */
    std::string toString();
    
    void setName(std::string name);
    
    static bool compare(Thread &t1, Thread &t2);
//     static Thread& compare(Thread *t1, Thread *t2);
    static bool init();
    static Thread& getCurrent();
    
  protected:
    Thread();
    ~Thread();
    
    /** setups thread function */
    virtual int pre_start();
    /** must implements thread code */
    virtual void* entry_point()=0 ;
    /** must sets the thread in non-running state, 
      here it should implements clean-up routine */
    virtual int exit_point();
    
    void *m_args ;
    void *m_ret_args ;
    pthread_t m_thread ;
  private:
    /** raw thread entry point, this function will call start_entry() */
    static void* _start(Thread* caller);
    
    bool m_isstarted ;
    bool m_isrunning ;
    bool m_iscancelled ;
    /** Name of thread */
    std::string m_name ;
//     static std::set<Thread&> m_threads ;
    static std::map<pthread_t,Thread*> m_runthreads ;
    static n2nc::sync::Mutex m_maplock ;
    
  /**
    \example threads_ex.cpp Threads example.
  */

};

/** \brief Workaround to emulate the main thread. Do not use this. */
/** Do not call this directly. Use Thread::init() instead.
    once Thread::init() is called by main thread, it 
    initializes a Thread object associated to main thread.
    By the way it is possible to call 
    eg. Thread::getCurrent()->setDescription("what im doing");
    and LOGMET() << "hello" ; output: Thread: what im doing Hello.
 */
class MainThread : public Thread {
  private:
    MainThread() : Thread() {
      /* NOTE Set the running state */
      this->pre_start();
      this->setName("Main");
    }
    ~MainThread(){}
    virtual void* entry_point(){}
//     virtual int pre_start(){}
//     virtual int exit_point(){}
  
  
  
  friend class Thread ;

};

/**
  \brief C-Style threading model
  
  This thread class utility take a function pointer instead implementing virtual method.
  \author fabsoft <fabsoft@gmail.com>
*/
class InlineThread : public Thread {
  public:
    /** Start routine fingerprint */
    typedef void *(*m_start_routine)(void*) ;
    typedef int (*m_stop_routine)(void*) ;
    /** Sets entry point for the thread */
    InlineThread(m_start_routine fnc,m_stop_routine fncstop = NULL) : Thread() {
      this->m_runfnc = fnc ;
      this->m_stopfnc = fncstop ;
    }
    ~InlineThread(){}
  private:
    virtual void* entry_point(){
      return this->m_runfnc(this->m_args);
    }
    virtual int exit_point(){
      if ( this->m_stopfnc == NULL )  return 0 ;
      return this->m_stopfnc(this->m_ret_args);
    }

  m_start_routine m_runfnc ;
  m_stop_routine m_stopfnc ;
};

}
}

#endif
